@@ -51,10 +51,10 @@ PyFortranObject_New(FortranDataDef* defs, f2py_void_func init) {
             PyDict_SetItemString(fp->dict,fp->defs[i].name,v);
         } else
             if ((fp->defs[i].data)!=NULL) { /* Is Fortran variable or array (not allocatable) */
-                if (fp->defs[i].type == PyArray_STRING) {
+                if (fp->defs[i].type == NPY_STRING) {
                     int n = fp->defs[i].rank-1;
                     v = PyArray_New(&PyArray_Type, n, fp->defs[i].dims.d,
-                                    PyArray_STRING, NULL, fp->defs[i].data, fp->defs[i].dims.d[n],
+                                    NPY_STRING, NULL, fp->defs[i].data, fp->defs[i].dims.d[n],
                                     NPY_FARRAY, NULL);
                 }
                 else {
@@ -92,55 +92,89 @@ fortran_dealloc(PyFortranObject *fp) {
 }
 
 
+#if PY_VERSION_HEX >= 0x03000000
+#else
 static PyMethodDef fortran_methods[] = {
     {NULL,          NULL}           /* sentinel */
 };
+#endif
 
 
 static PyObject *
 fortran_doc (FortranDataDef def) {
     char *p;
+    /*
+      p is used as a buffer to hold generated documentation strings.
+      A common operation in generating the documentation strings, is
+      appending a string to the buffer p. Earlier, the following
+      idiom was:
+
+        sprintf(p, "%s<string to be appended>", p);
+
+      but this does not work when _FORTIFY_SOURCE=2 is enabled: instead
+      of appending the string, the string is inserted.
+
+      As a fix, the following idiom should be used for appending
+      strings to a buffer p:
+
+        sprintf(p + strlen(p), "<string to be appended>");
+     */
     PyObject *s = NULL;
     int i;
     unsigned size=100;
     if (def.doc!=NULL)
         size += strlen(def.doc);
     p = (char*)malloc (size);
+    p[0] = '\0'; /* make sure that the buffer has zero length */
     if (sprintf(p,"%s - ",def.name)==0) goto fail;
     if (def.rank==-1) {
         if (def.doc==NULL) {
-            if (sprintf(p,"%sno docs available",p)==0)
+            if (sprintf(p+strlen(p),"no docs available")==0)
                 goto fail;
         } else {
-            if (sprintf(p,"%s%s",p,def.doc)==0)
+            if (sprintf(p+strlen(p),"%s",def.doc)==0)
                 goto fail;
         }
     } else {
         PyArray_Descr *d = PyArray_DescrFromType(def.type);
-        if (sprintf(p,"%s'%c'-",p,d->type)==0) {Py_DECREF(d); goto fail;}
+        if (sprintf(p+strlen(p),"'%c'-",d->type)==0) {
+	  Py_DECREF(d);
+	  goto fail;
+	}
         Py_DECREF(d);
         if (def.data==NULL) {
-            if (sprintf(p,"%sarray(%" NPY_INTP_FMT,p,def.dims.d[0])==0) goto fail;
+            if (sprintf(p+strlen(p),"array(%" NPY_INTP_FMT,def.dims.d[0])==0)
+	      goto fail;
             for(i=1;i<def.rank;++i)
-                if (sprintf(p,"%s,%" NPY_INTP_FMT,p,def.dims.d[i])==0) goto fail;
-            if (sprintf(p,"%s), not allocated",p)==0) goto fail;
+                if (sprintf(p+strlen(p),",%" NPY_INTP_FMT,def.dims.d[i])==0)
+		  goto fail;
+            if (sprintf(p+strlen(p),"), not allocated")==0)
+	      goto fail;
         } else {
             if (def.rank>0) {
-                if (sprintf(p,"%sarray(%"NPY_INTP_FMT,p,def.dims.d[0])==0) goto fail;
+                if (sprintf(p+strlen(p),"array(%"NPY_INTP_FMT,def.dims.d[0])==0)
+		  goto fail;
                 for(i=1;i<def.rank;i++)
-                    if (sprintf(p,"%s,%" NPY_INTP_FMT,p,def.dims.d[i])==0) goto fail;
-                if (sprintf(p,"%s)",p)==0) goto fail;
+                    if (sprintf(p+strlen(p),",%" NPY_INTP_FMT,def.dims.d[i])==0)
+		      goto fail;
+                if (sprintf(p+strlen(p),")")==0) goto fail;
             } else {
-                if (sprintf(p,"%sscalar",p)==0) goto fail;
+                if (sprintf(p+strlen(p),"scalar")==0) goto fail;
             }
         }
     }
-    if (sprintf(p,"%s\n",p)==0) goto fail;
+    if (sprintf(p+strlen(p),"\n")==0) goto fail;
     if (strlen(p)>size) {
-        fprintf(stderr,"fortranobject.c:fortran_doc:len(p)=%zd>%d(size): too long doc string required, increase size\n",strlen(p),size);
+        fprintf(stderr,"fortranobject.c:fortran_doc:len(p)=%zd>%d(size):"\
+		" too long doc string required, increase size\n",\
+		strlen(p),size);
         goto fail;
     }
+#if PY_VERSION_HEX >= 0x03000000
+    s = PyUnicode_FromString(p);
+#else
     s = PyString_FromString(p);
+#endif
  fail:
     free(p);
     return s;
@@ -194,20 +228,41 @@ fortran_getattr(PyFortranObject *fp, char *name) {
         return fp->dict;
     }
     if (strcmp(name,"__doc__")==0) {
+#if PY_VERSION_HEX >= 0x03000000
+        PyObject *s = PyUnicode_FromString(""), *s2, *s3;
+        for (i=0;i<fp->len;i++) {
+            s2 = fortran_doc(fp->defs[i]);
+            s3 = PyUnicode_Concat(s, s2);
+            Py_DECREF(s2);
+            Py_DECREF(s);
+            s = s3;
+        }
+#else
         PyObject *s = PyString_FromString("");
         for (i=0;i<fp->len;i++)
             PyString_ConcatAndDel(&s,fortran_doc(fp->defs[i]));
+#endif
         if (PyDict_SetItemString(fp->dict, name, s))
             return NULL;
         return s;
     }
     if ((strcmp(name,"_cpointer")==0) && (fp->len==1)) {
-        PyObject *cobj = PyCObject_FromVoidPtr((void *)(fp->defs[0].data),NULL);
+        PyObject *cobj = F2PyCapsule_FromVoidPtr((void *)(fp->defs[0].data),NULL);
         if (PyDict_SetItemString(fp->dict, name, cobj))
             return NULL;
         return cobj;
     }
+#if PY_VERSION_HEX >= 0x03000000
+    if (1) {
+        PyObject *str, *ret;
+        str = PyUnicode_FromString(name);
+        ret = PyObject_GenericGetAttr((PyObject *)fp, str);
+        Py_DECREF(str);
+        return ret;
+    }
+#else
     return Py_FindMethod(fortran_methods, (PyObject *)fp, name);
+#endif
 }
 
 static int
@@ -295,10 +350,39 @@ fortran_call(PyFortranObject *fp, PyObject *arg, PyObject *kw) {
     return NULL;
 }
 
+static PyObject *
+fortran_repr(PyFortranObject *fp)
+{
+    PyObject *name = NULL, *repr = NULL;
+    name = PyObject_GetAttrString((PyObject *)fp, "__name__");
+    PyErr_Clear();
+#if PY_VERSION_HEX >= 0x03000000
+    if (name != NULL && PyUnicode_Check(name)) {
+        repr = PyUnicode_FromFormat("<fortran %U>", name);
+    }
+    else {
+        repr = PyUnicode_FromString("<fortran object>");
+    }
+#else
+    if (name != NULL && PyString_Check(name)) {
+        repr = PyString_FromFormat("<fortran %s>", PyString_AsString(name));
+    }
+    else {
+        repr = PyString_FromString("<fortran object>");
+    }
+#endif
+    Py_XDECREF(name);
+    return repr;
+}
+
 
 PyTypeObject PyFortran_Type = {
+#if PY_VERSION_HEX >= 0x03000000
+    PyVarObject_HEAD_INIT(NULL, 0)
+#else
     PyObject_HEAD_INIT(0)
     0,                    /*ob_size*/
+#endif
     "fortran",                    /*tp_name*/
     sizeof(PyFortranObject),      /*tp_basicsize*/
     0,                    /*tp_itemsize*/
@@ -307,8 +391,8 @@ PyTypeObject PyFortran_Type = {
     0,                    /*tp_print*/
     (getattrfunc)fortran_getattr, /*tp_getattr*/
     (setattrfunc)fortran_setattr, /*tp_setattr*/
-    0,                    /*tp_compare*/
-    0,                    /*tp_repr*/
+    0,                    /*tp_compare/tp_reserved*/
+    (reprfunc)fortran_repr, /*tp_repr*/
     0,                    /*tp_as_number*/
     0,                    /*tp_as_sequence*/
     0,                    /*tp_as_mapping*/
@@ -502,6 +586,7 @@ static int swap_arrays(PyArrayObject* arr1, PyArrayObject* arr2) {
     (  (PyArray_ISINTEGER(arr) && PyTypeNum_ISINTEGER(type_num))        \
        ||(PyArray_ISFLOAT(arr) && PyTypeNum_ISFLOAT(type_num))          \
        ||(PyArray_ISCOMPLEX(arr) && PyTypeNum_ISCOMPLEX(type_num))      \
+       ||(PyArray_ISBOOL(arr) && PyTypeNum_ISBOOL(type_num))            \
        )
 
 extern
@@ -587,10 +672,17 @@ PyArrayObject* array_from_pyobj(const int type_num,
         if (check_and_fix_dimensions(arr,rank,dims)) {
             return NULL; /*XXX: set exception */
         }
-
+	/*
+	printf("intent alignement=%d\n", F2PY_GET_ALIGNMENT(intent));
+	printf("alignement check=%d\n", F2PY_CHECK_ALIGNMENT(arr, intent));
+	int i;
+	for (i=1;i<=16;i++)
+	  printf("i=%d isaligned=%d\n", i, ARRAY_ISALIGNED(arr, i));
+	*/
         if ((! (intent & F2PY_INTENT_COPY))
             && PyArray_ITEMSIZE(arr)==elsize
             && ARRAY_ISCOMPATIBLE(arr,type_num)
+	    && F2PY_CHECK_ALIGNMENT(arr, intent)
             ) {
             if ((intent & F2PY_INTENT_C)?PyArray_ISCARRAY(arr):PyArray_ISFARRAY(arr)) {
                 if ((intent & F2PY_INTENT_OUT)) {
@@ -615,6 +707,8 @@ PyArrayObject* array_from_pyobj(const int type_num,
             if (!(ARRAY_ISCOMPATIBLE(arr,type_num)))
                 sprintf(mess+strlen(mess)," -- input '%c' not compatible to '%c'",
                         arr->descr->type,typechar);
+	    if (!(F2PY_CHECK_ALIGNMENT(arr, intent)))
+	      sprintf(mess+strlen(mess)," -- input not %d-aligned", F2PY_GET_ALIGNMENT(intent));
             PyErr_SetString(PyExc_ValueError,mess);
             return NULL;
         }
@@ -694,18 +788,20 @@ int check_and_fix_dimensions(const PyArrayObject* arr,const int rank,npy_intp *d
         npy_intp new_size = 1;
         int free_axe = -1;
         int i;
+        npy_intp d;
         /* Fill dims where -1 or 0; check dimensions; calc new_size; */
         for(i=0;i<arr->nd;++i) {
+            d = arr->dimensions[i];
             if (dims[i] >= 0) {
-                if (dims[i]!=arr->dimensions[i]) {
+                if (d>1 && dims[i]!=d) {
                     fprintf(stderr,"%d-th dimension must be fixed to %" NPY_INTP_FMT
                             " but got %" NPY_INTP_FMT "\n",
-                            i,dims[i], arr->dimensions[i]);
+                            i,dims[i], d);
                     return 1;
                 }
                 if (!dims[i]) dims[i] = 1;
             } else {
-                dims[i] = arr->dimensions[i] ? arr->dimensions[i] : 1;
+                dims[i] = d ? d : 1;
             }
             new_size *= dims[i];
         }
@@ -724,16 +820,17 @@ int check_and_fix_dimensions(const PyArrayObject* arr,const int rank,npy_intp *d
             new_size *= dims[free_axe];
         }
         if (new_size != arr_size) {
-            fprintf(stderr,"confused: new_size=%" NPY_INTP_FMT
-                    ", arr_size=%" NPY_INTP_FMT " (maybe too many free"
+            fprintf(stderr,"unexpected array size: new_size=%" NPY_INTP_FMT
+                    ", got array with arr_size=%" NPY_INTP_FMT " (maybe too many free"
                     " indices)\n", new_size,arr_size);
             return 1;
         }
     } else if (rank==arr->nd) {
+        npy_intp new_size = 1;
         int i;
         npy_intp d;
         for (i=0; i<rank; ++i) {
-            d = arr->dimensions[i];
+	    d = arr->dimensions[i];
             if (dims[i]>=0) {
                 if (d > 1 && d!=dims[i]) {
                     fprintf(stderr,"%d-th dimension must be fixed to %" NPY_INTP_FMT
@@ -743,6 +840,12 @@ int check_and_fix_dimensions(const PyArrayObject* arr,const int rank,npy_intp *d
                 }
                 if (!dims[i]) dims[i] = 1;
             } else dims[i] = d;
+            new_size *= dims[i];
+        }
+        if (new_size != arr_size) {
+            fprintf(stderr,"unexpected array size: new_size=%" NPY_INTP_FMT
+                    ", got array with arr_size=%" NPY_INTP_FMT "\n", new_size,arr_size);
+            return 1;
         }
     } else { /* [[1,2]] -> [[1],[2]] */
         int i,j;
@@ -782,7 +885,7 @@ int check_and_fix_dimensions(const PyArrayObject* arr,const int rank,npy_intp *d
         }
         for (i=0,size=1;i<rank;++i) size *= dims[i];
         if (size != arr_size) {
-            fprintf(stderr,"confused: size=%" NPY_INTP_FMT ", arr_size=%" NPY_INTP_FMT
+            fprintf(stderr,"unexpected array size: size=%" NPY_INTP_FMT ", arr_size=%" NPY_INTP_FMT
                     ", rank=%d, effrank=%d, arr.nd=%d, dims=[",
                     size,arr_size,rank,effrank,arr->nd);
             for (i=0;i<rank;++i) fprintf(stderr," %" NPY_INTP_FMT,dims[i]);
@@ -810,6 +913,61 @@ int copy_ND_array(const PyArrayObject *arr, PyArrayObject *out)
     return PyArray_CopyInto(out, (PyArrayObject *)arr);
 }
 
+/*********************************************/
+/* Compatibility functions for Python >= 3.0 */
+/*********************************************/
+
+#if PY_VERSION_HEX >= 0x03000000
+
+PyObject *
+F2PyCapsule_FromVoidPtr(void *ptr, void (*dtor)(PyObject *))
+{
+    PyObject *ret = PyCapsule_New(ptr, NULL, dtor);
+    if (ret == NULL) {
+        PyErr_Clear();
+    }
+    return ret;
+}
+
+void *
+F2PyCapsule_AsVoidPtr(PyObject *obj)
+{
+    void *ret = PyCapsule_GetPointer(obj, NULL);
+    if (ret == NULL) {
+        PyErr_Clear();
+    }
+    return ret;
+}
+
+int
+F2PyCapsule_Check(PyObject *ptr)
+{
+    return PyCapsule_CheckExact(ptr);
+}
+
+#else
+
+PyObject *
+F2PyCapsule_FromVoidPtr(void *ptr, void (*dtor)(void *))
+{
+    return PyCObject_FromVoidPtr(ptr, dtor);
+}
+
+void *
+F2PyCapsule_AsVoidPtr(PyObject *ptr)
+{
+    return PyCObject_AsVoidPtr(ptr);
+}
+
+int
+F2PyCapsule_Check(PyObject *ptr)
+{
+    return PyCObject_Check(ptr);
+}
+
+#endif
+
+
 #ifdef __cplusplus
 }
 #endif
